/*
 * This file is part of libphidget22java
 *
 * Copyright 2015 Phidgets Inc <patrick@phidgets.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see
 * <http://www.gnu.org/licenses/>
 */

#ifndef PHIDGET_JNI_H
#define PHIDGET_JNI_H

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>

#include "jni.h"

#include "phidget22.h"
#include "phidget_enums_jni.h"

 /* Define here structs which are not in the library */
typedef struct {
	const char *code;
	int bitCount;
} IRCode, *IRCodeHandle;

typedef struct {
	const char *code;
	PhidgetIR_CodeInfo codeInfo;
} IRLearnedCode, *IRLearnedCodeHandle;

typedef struct {
	const char *tagString;
	PhidgetRFID_Protocol protocol;
} RFIDTag, *RFIDTagHandle;

typedef struct {
	uint64_t size;
	int keepCount;
} LogRotating, *LogRotatingHandle;

typedef struct {
	uint32_t distances[8];
	uint32_t amplitudes[8];
	uint32_t count;
} DistanceSensorSonarReflections, *DistanceSensorSonarReflectionsHandle;

#include "phidget_structs_jni.h"

extern const char * CCONV Phidget_strerror(PhidgetReturnCode error);
extern void asyncCallbackHandler(PhidgetHandle phid, void *ctx, PhidgetReturnCode returnCode);

void CCONV setOnJavaDetachCurrentThreadHandler(void(CCONV *fptr)(void));

extern void com_phidget22_OnLoad(JNIEnv *env);

#ifdef _ANDROID
#define JNIEnvPtr const struct JNINativeInterface ***
#else
#define JNIEnvPtr void **
#endif

extern JavaVM *ph_vm;

extern jclass ph_exception_class;
extern jmethodID ph_exception_cons1;
extern jmethodID ph_exception_cons2;
extern jclass phid_class;
extern jmethodID phid_cons;
extern jfieldID phidget_handle_fid;

#ifdef _ANDROID
extern int AndroidUsbJarAvailable;
int com_phidget22_usb_Manager_OnLoad(JNIEnv *);
int com_phidget22_usb_Phidget_OnLoad(JNIEnv *);
#endif

jlong updateGlobalRef(JNIEnv *env, jobject obj, jfieldID fid, jboolean b);

#define JNI_ABORT_STDERR(...) \
do { \
	fprintf(stderr, "\nAn error was encountered in the phidget22java jni library. This program will now be aborted.\n"); \
	fprintf(stderr, "Make sure phidget22.jar and phidget22 C libraries are in sync.\n\n"); \
	fprintf(stderr, "Error Details:\n"); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n\n"); \
	(*env)->ExceptionDescribe(env); \
	(*env)->ExceptionClear(env); \
	abort(); \
} while(0)

#define PH_THROW_LAST_ERR(errno) do { \
	jobject eobj; \
	jstring edesc; \
	jstring edetail; \
	const char *errorDesc; \
	char *errorDetail; \
	size_t errorDetailLen; \
	PhidgetReturnCode code; \
 \
	if(Phidget_getLastError(&code, &errorDesc, NULL, &errorDetailLen) == EPHIDGET_OK && code == errno) { \
		errorDetail = malloc(errorDetailLen); \
		Phidget_getLastError(&code, &errorDesc, errorDetail, &errorDetailLen); \
		if (!(edesc = (*env)->NewStringUTF(env, errorDesc))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(edetail = (*env)->NewStringUTF(env, errorDetail))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons1, code, edesc, edetail))) \
			JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
		(*env)->DeleteLocalRef (env, edetail); \
		free(errorDetail); \
	} else { \
		if (!(edesc = (*env)->NewStringUTF(env, Phidget_strerror(errno)))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons2, errno, edesc))) \
			JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
	} \
	(*env)->DeleteLocalRef (env, edesc); \
	(*env)->Throw(env, (jthrowable)eobj); \
	(*env)->DeleteLocalRef (env, eobj); \
} while(0)

#define PH_THROW(errno) do { \
	jobject eobj; \
	jstring edesc; \
 \
	if (!(edesc = (*env)->NewStringUTF(env, Phidget_strerror(errno)))) \
		JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
	if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons2, errno, edesc))) \
		JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
	(*env)->DeleteLocalRef (env, edesc); \
	(*env)->Throw(env, (jthrowable)eobj); \
	(*env)->DeleteLocalRef (env, eobj); \
} while(0)


#endif
